from pydantic import BaseModel, FilePath, validator
from functools import cached_property
from jinja2 import Environment, BaseLoader, TemplateSyntaxError 
from typing import Union, AbstractSet, Mapping, Any
from .validated_cached_property import ValidatedCachedProperty

AbstractSetIntStr = Union[AbstractSet, int, str]
MappingIntStrAny = Union[Mapping, int, str, Any]
DictStrAny = Union[dict, str, Any]

class Phraction(BaseModel):
    """The base unit of construction of a Phractal document - analagous to React.Component.
    Renders a supplied template as an HTML string, automatically inserting field values.
    
    Performs Pydantic-based field validation/type-checking. 
    
    Args:
        template: A Jinja2-formatted string defining the HTML structure of the Phraction.
    """

    template: str

    class Config:
        arbitrary_types_allowed = True
        validate_assignment = True
        keep_untouched = (cached_property,)

    def __repr__(self):
        template = Environment(loader=BaseLoader).from_string(self.template)
        return template.render(self.dict())

    def __str__(self):
        return repr(self)

    def save(self, path: FilePath):
        """Writes a rendered Phraction document to file as HTML.

        Args:
            path: The local filepath to which the rendered document should be saved.        
        """
        with open(path, 'w') as output_file:
                output_file.write(repr(self))

    @validator("template", always=True)
    @classmethod
    def check_template(cls, value) -> str:
        
        if not isinstance(value, str):
            raise ValueError(f"Invalid template supplied to {cls.__name__}. Template must be a Jinja2-formatted string.")
        
        try:
            env = Environment(loader=BaseLoader)
            env.parse(value)
        except TemplateSyntaxError as e:
            raise ValueError(f"Invalid template supplied to {cls.__name__}. Error on line {e.lineno} of template.\n\tError message: {e.message}")
        return value

    @classmethod
    def get_properties(cls) -> list[str]:
        return [prop for prop in dir(cls) if isinstance(getattr(cls, prop), property) or isinstance(getattr(cls, prop), ValidatedCachedProperty)]

    def dict(
        self,
        *,
        include: Union['AbstractSetIntStr', 'MappingIntStrAny'] = None,
        exclude: Union['AbstractSetIntStr', 'MappingIntStrAny'] = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> 'DictStrAny':
        """Performs identically to the Pydantic Basemodel's .dict() method, except that Properties and ValidatedCachedProperties are also included."""

        attribs = super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none
        )
        props = self.get_properties()
        # Include and exclude properties
        if include:
            props = [prop for prop in props if prop in include]
        if exclude:
            props = [prop for prop in props if prop not in exclude]

        # Update the attribute dict with the properties
        if props:
            attribs.update({prop: getattr(self, prop) for prop in props})

        return attribs

    def with_boilerplate(self, bootstrap: bool=False, title: str="Phractal Document"):
        """A method to wrap the output of a phraction in standard HTML boilerplate. Use ONLY on the top-level Phraction in a doc.

        args:
            bootstrap: Boolean. Indicates whether or not to include Bootstrap styling script v4.3.1 in head.
            title: String. The <title> to assign in the document's head. Default is "Phractal Document".
        
        returns:
            An object of the Phraction class which wraps the calling instance in boilerplate.
        """
        class Boilerplater(Phraction):
            template = """
            <!DOCTYPE html>
            <html lang="en">
            <head>
                <meta charset="UTF-8">
                <meta name="viewport" content="width=device-width, initial-scale=1.0">
                <meta http-equiv="X-UA-Compatible" content="ie=edge">
                {% if bootstrap %}
                    <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@4.3.1/dist/css/bootstrap.min.css" integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T" crossorigin="anonymous">
                {% endif %}
                <title>{{ title }}</title>
            </head>
            <body>
                {{ inner_phractal }}
            </body>
            </html>"""
            title: str
            bootstrap: bool
            
            @property
            def inner_phractal(obj):
                return self

        boilerplater=Boilerplater(
            title = title,
            bootstrap = bootstrap
        )
        
        return boilerplater



import inspect

def _get_handy_context(func: callable) -> tuple[str, int, str]:
    """Grabs the filename, line number, and code-block in which a function is defined.
    
    Args:
        func - the function being interrogated.
    """
    filename = inspect.getfile(func)
    code, lineno = inspect.getsourcelines(func)
    context = "".join(code)

    return filename, lineno, context


import unittest
from src.phraction import Phraction
from src.validated_cached_property import ValidatedCachedProperty
from pydantic import ValidationError

class TestTemplate(unittest.TestCase):
    def test_correct_template_validation(self):
        try:
            class GoodPhraction(Phraction):
                template = "<p>{{ msg }}</p>"
            hi = GoodPhraction()
        except ValidationError as e:
            self.fail("A Phraction with a correct template raised a validation error.")

    def test_incorrect_template_validation(self):
        with self.assertRaises(
            ValidationError, 
            msg="A malformed template was provided but no exception was raised."
        ):
            class BadPhraction(Phraction):
                template = "<p>{{ msg }</p>"
                msg: str

            hi = BadPhraction(msg="hi")

class TestRender(unittest.TestCase):
    def test_blank_template_render(self):
        class SomePhraction(Phraction):
            template = "<p>hi</p>"
        hi = SomePhraction()
        self.assertEqual(str(hi), "<p>hi</p>")

    def test_template_variable_render(self):
        class SomePhraction(Phraction):
            template = "<p>{{ somevar }}</p>"
            somevar: str
        hi = SomePhraction(somevar="hi")
        self.assertEqual(str(hi), "<p>hi</p>")

class TestValidation(unittest.TestCase):
    def test_validate_incorrect_type_variable(self):
        class SomePhraction(Phraction):
            template = "<p>{{ somevar }}</p>"
            somevar: int
        
        with self.assertRaises(
            ValidationError, 
            msg="Incorrect type did not raise a ValidationError."
        ):
            hi = SomePhraction(somevar="hi")

    def test_validate_correct_type_variable(self):
        class SomePhraction(Phraction):
            template = "<p>{{ somevar }}</p>"
            somevar: int
        
        try:
            hi = SomePhraction(somevar=1)
        except ValidationError as e:
            self.fail("Correctly-typed variable raised a validation error.")

class TestProperties(unittest.TestCase):
    def test_properties_included_in_render(self):
        class SomePhraction(Phraction):
            template = "<p>{{ somevar }}</p>"

            @property
            def somevar(self):
                return "hi"
            
        self.assertEqual(str(SomePhraction()), "<p>hi</p>")

    def test_validated_cached_properties_included_in_render(self):
        class SomePhraction(Phraction):
            template = "<p>{{ somevar }}</p>"

            @ValidatedCachedProperty
            def somevar(self) -> str:
                return "hi"
            
        self.assertEqual(str(SomePhraction()), "<p>hi</p>")

class TestNesting(unittest.TestCase):
    def test_nesting(self):
        class FirstPhrac(Phraction):
            template = "{{ some_var }}"
            some_var: str

        class SecondPhrac(Phraction):
            template = "<p>{{ first_phrac }}</p>"
            some_var: str

            @property
            def first_phrac(self):
                return FirstPhrac(some_var=self.some_var)
            
        test_phrac = SecondPhrac(some_var="hi")
        
        self.assertEqual(str(test_phrac), "<p>hi</p>")

class TestBoilerplate(unittest.TestCase):
    def test_output_is_phraction(self):
        
        class MyDoc(Phraction):
            template="<p>{{ msg }}</p>"
            msg: str

        my_doc = MyDoc(msg="Hi!")
        self.assertIsInstance(my_doc.with_boilerplate(), Phraction)

    def test_boilerplate_is_correct(self):
        output = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <meta http-equiv="X-UA-Compatible" content="ie=edge">
            <title>Phractal Document</title>
        </head>
        <body>
            <p>Hi!</p>
        </body>
        </html>"""
        
        class MyDoc(Phraction):
            template="<p>{{ msg }}</p>"
            msg: str

        my_doc = MyDoc(msg="Hi!")
        self.assertEqual("".join(str(my_doc.with_boilerplate()).split()), "".join(output.split()))

    def test_bootstrap_gets_included(self):
        output = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <meta http-equiv="X-UA-Compatible" content="ie=edge">
            <link rel="stylesheet" href="https://cdn.jsdelivr.net/npm/bootstrap@4.3.1/dist/css/bootstrap.min.css" integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T" crossorigin="anonymous">
            <title>Phractal Document</title>
        </head>
        <body>
            <p>Hi!</p>
        </body>
        </html>"""
        
        class MyDoc(Phraction):
            template="<p>{{ msg }}</p>"
            msg: str

        my_doc = MyDoc(msg="Hi!")
        self.assertEqual("".join(str(my_doc.with_boilerplate(bootstrap=True)).split()), "".join(output.split()))

    def test_title_correct(self):
        output = """
        <!DOCTYPE html>
        <html lang="en">
        <head>
            <meta charset="UTF-8">
            <meta name="viewport" content="width=device-width, initial-scale=1.0">
            <meta http-equiv="X-UA-Compatible" content="ie=edge">
            <title>A Snappy Title</title>
        </head>
        <body>
            <p>Hi!</p>
        </body>
        </html>"""
        
        class MyDoc(Phraction):
            template="<p>{{ msg }}</p>"
            msg: str

        my_doc = MyDoc(msg="Hi!")
        self.assertEqual("".join(str(my_doc.with_boilerplate(title="A Snappy Title")).split()), "".join(output.split()))


from functools import cached_property
from pydantic import BaseModel, ValidationError
from ._utils import _get_handy_context


_NOT_FOUND = object()

class ValidatedCachedProperty(cached_property):
    """A decorator which turns a function into a property on a class which is cached and validated. 
    
    Validation is based on return type hints (type hints are compulsory).
    Does not currently support parameterised generics (ex: `list[int]` is not a valid type hint - use `list` instead) -> This is an open issue!
    Value is calculated and cached at first access.
    Subclassed from the inbuilt functools.cached_property.      
    """

    def _validate_return_type(self, val):
        """Validates the return type of the property based on type hints.
        """

        if not 'return' in self.func.__annotations__:
            filename, lineno, context = _get_handy_context(self.func)
            raise ValueError(f"No type hint supplied for ValidatedCachedProperty {self.attrname}. Context: \n {filename} ({lineno}): \n {context}")
        
        try:
            class ValidationProxy(BaseModel):
                class Config:
                    arbitrary_types_allowed = True
                returnvar: self.func.__annotations__['return']
            
            validation_proxy = ValidationProxy(returnvar=val)

        except ValidationError as e:
            filename, lineno, context = _get_handy_context(self.func)
            raise ValueError(f"ValidatedCachedProperty {self.attrname} returned a value that did not match its type hint. Hinted type: {self.func.__annotations__['return']}. Returned type: {type(val)}. Context: \n {filename} ({lineno}): \n {context}") 
                
    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        if self.attrname is None:
            raise TypeError(
                "Cannot use cached_property instance without calling __set_name__ on it.")
        try:
            cache = instance.__dict__
        except AttributeError:  # not all objects have __dict__ (e.g. class defines slots)
            msg = (
                f"No '__dict__' attribute on {type(instance).__name__!r} "
                f"instance to cache {self.attrname!r} property."
            )
            raise TypeError(msg) from None
        val = cache.get(self.attrname, _NOT_FOUND)
        if val is _NOT_FOUND:
            with self.lock:
                # check if another thread filled cache while we awaited lock
                val = cache.get(self.attrname, _NOT_FOUND)
                if val is _NOT_FOUND:
                    val = self.func(instance)
                    try:
                        cache[self.attrname] = val
                    except TypeError:
                        msg = (
                            f"The '__dict__' attribute on {type(instance).__name__!r} instance "
                            f"does not support item assignment for caching {self.attrname!r} property."
                        )
                        raise TypeError(msg) from None
        
        self._validate_return_type(val)            
        
        return val
    


import unittest
from typing import Any
from src.validated_cached_property import ValidatedCachedProperty
from pydantic import ValidationError

class TestVCP(unittest.TestCase):
    def test_correct_hint_validation(self):
        
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self) -> int:
                return 1
        try:
            test_thing = TestClass()
            test_thing.test_prop
        except (ValidationError, ValueError) as e:
            self.fail("A VCP with correct type-hinting incorrectly raised an error.")

    def test_incorrect_hint_validation(self):
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self) -> int:
                return "hi"
        with self.assertRaises(
            ValueError,
            msg="An incorrectly typed type hint did not raise a ValueError."
        ):
            test_thing = TestClass()
            test_thing.test_prop
        
    def test_no_type_hint(self):
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self):
                return "hi"
        with self.assertRaises(
            ValueError,
            msg="A missing type hint did not raise a ValueError."
        ):
            test_thing = TestClass()
            test_thing.test_prop

    def test_correct_parameterized_generic(self):
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self) -> list[int]:
                return [1, 2, 3]

        try:
            test_thing = TestClass()
            test_thing.test_prop
        except (ValidationError, ValueError) as e:
            self.fail("A VCP with correct type-hinting (list[int]) incorrectly raised an error.")

    def test_union_return_type_hint(self):
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self) -> int|bool:
                return 3
            
        try:
            test_thing = TestClass()
            test_thing.test_prop
        except (ValidationError, ValueError) as e:
            self.fail("A VCP with correct type-hinting (int|bool) incorrectly raised an error.")

    def test_any_return_hint(self):
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self) -> Any:
                return 3
            
        try:
            test_thing = TestClass()
            test_thing.test_prop
        except (ValidationError, ValueError) as e:
            self.fail("A VCP with correct type-hinting (Any) incorrectly raised an error.")

    def test_incorrect_parameterized_generic(self):
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self) -> list[int]:
                return [1, "b", 3]
            
        with self.assertRaises(ValueError) as e:
            test_thing = TestClass()
            test_thing.test_prop

    def test_incorrect_union_hint(self):
        class TestClass:
            @ValidatedCachedProperty
            def test_prop(self) -> int|bool:
                return "hi"
            
        with self.assertRaises(ValueError) as e:
            test_thing = TestClass()
            test_thing.test_prop

    def test_parameterized_non_generics(self):
        class TestClass1:
            pass

        class TestClass2:
            @ValidatedCachedProperty
            def test_prop(self) -> list[TestClass1]:
                return [TestClass1(), TestClass1()]
            
        try:
            test_thing = TestClass2()
            test_thing.test_prop
        except (ValidationError, ValueError) as e:
            self.fail("A VCP with correct type-hinting (Any) incorrectly raised an error.")

    def test_incorrect_parameterized_non_generics(self):
        class TestClass1:
            pass

        class TestClass3:
            pass

        class TestClass2:
            @ValidatedCachedProperty
            def test_prop(self) -> list[TestClass1]:
                return [TestClass1(), TestClass3()]
            
        with self.assertRaises(ValueError) as e:
            test_thing = TestClass2()
            test_thing.test_prop


from setuptools import find_packages, setup

with open("./README.md", "r") as f:
    long_description = f.read()

setup(
    name="Phractal",
    version="0.0.3a",
    description="A tool for creating HTML documents programmatically with Python",
    package_dir={"": "app"},
    packages=find_packages(where="app"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Hauteclere/phractal",
    author="Hauteclere",
    author_email="hauteclere.code@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "Jinja2 >= 3.1.2", 
        "MarkupSafe >=2.1.2", 
        "pydantic >= 1.10.6", 
        "typing_extensions >= 4.5.0"
    ],
    extras_require={
        "dev": ["twine>=4.0.2"],
    },
    python_requires=">=3.10",
)
